#include "TCP.h"
#include "utility/w5500.h"
#include "utility/socket.h"

EthernetMaketClient::EthernetMaketClient() {
    _sock = MAX_SOCK_NUM;
}

EthernetMaketClient::EthernetMaketClient(uint8_t sock) {
    _sock = sock;
}

EthernetMaketClient::EthernetMaketClient(EthernetClient &client) {
    _sock = client.getSocketNumber();
}

EthernetMaketClient& EthernetMaketClient::operator=(const EthernetMaketClient& other) {
    if (this == &other) {
        return *this; // Защита от самоприсваивания
    }

    this->lastActivityTime = other.lastActivityTime;
    this->startConnectionTime = other.startConnectionTime;
    this->stopStartTime = other.stopStartTime;
    this->isNonBlocking = other.isNonBlocking;
    this->connectStatus = other.connectStatus;
    this->dstIP = other.dstIP;
    this->dstPort = other.dstPort;
    this->dataPtr = other.dataPtr;
    this->dataSize = other.dataSize;
    this->onSendSuccess = other.onSendSuccess;
    this->_sock = other._sock;

    const_cast<EthernetMaketClient&>(other).connectStatus = CONNECT_IDLE;

    return *this;
}

ConnectionStatusSimple EthernetMaketClient::connectNonBlock(IPAddress ip, uint16_t port) {
    dstIP = ip;
    dstPort = port;

    if (_sock != MAX_SOCK_NUM) {
        uint8_t status_ = status();
        switch (status_) {
            case SnSR::ESTABLISHED:
                connectStatus = CONNECT_SUCCESS;
                return CONNECT_SUCCESS;
            case SnSR::CLOSED:
            case SnSR::CLOSE_WAIT:
                close(); // Освобождаем сокет перед попыткой нового подключения
                break;
            default:
                return CONNECT_CONNECTING;
        }
    }

    for (int i = 0; i < MAX_SOCK_NUM; i++) {
        uint8_t s = w5500.readSnSR(i);
        if (s == SnSR::CLOSED || s == SnSR::CLOSE_WAIT) {
            _sock = i;
            break;
        }
    }

    if (_sock == MAX_SOCK_NUM) {
        connectStatus = CONNECT_FAIL;
        return CONNECT_FAIL;
    }

    _srcport++;
    if (_srcport == 0) _srcport = 1024;
    socket(_sock, SnMR::TCP, _srcport, 0);

    if (!::connect(_sock, rawIPAddress(ip), port)) {
        _sock = MAX_SOCK_NUM;
        connectStatus = CONNECT_FAIL;
        return CONNECT_FAIL;
    }

    connectStatus = CONNECT_CONNECTING;
    return CONNECT_CONNECTING;
}

void EthernetMaketClient::tick_() {
    if (_sock == MAX_SOCK_NUM) {
        connectStatus = CONNECT_IDLE;
        return;
    }

    uint8_t s = w5500.readSnSR(_sock);

    switch (s) {
        case SnSR::CLOSED:
            connectStatus = (connectStatus == CONNECT_CONNECTING) ? CONNECT_FAIL : CONNECT_STOP_WAITING;
            close();
            break;
        case SnSR::ESTABLISHED:
            if (connectStatus != CONNECT_CONNECTED) {
                connectStatus = CONNECT_SUCCESS;
            }
            break;
        case SnSR::CLOSE_WAIT:
            connectStatus = CONNECT_STOP_START;
            disconnect();
            break;
        case SnSR::FIN_WAIT:
        case SnSR::CLOSING:
        case SnSR::TIME_WAIT:
        case SnSR::LAST_ACK:
            close();
            connectStatus = CONNECT_STOP_WAITING;
            break;
        default:
            break;
    }

    // Проверка активности и таймаутов
    if (connectStatus == CONNECT_CONNECTED) {
        if (millis() - lastActivityTime > timeout) {
            connectStatus = CONNECT_STOP_START;
            disconnect();
        }
    } else if (connectStatus == CONNECT_CONNECTING) {
        if (millis() - startConnectionTime >= connectionTimeout) {
            connectStatus = CONNECT_FAIL;
            close();
        }
    }
}

void EthernetMaketClient::disconnect() {
    if (_sock != MAX_SOCK_NUM) {
        ::disconnect(_sock);
    }
}

void EthernetMaketClient::close() {
    if (_sock != MAX_SOCK_NUM) {
        ::close(_sock);
        _sock = MAX_SOCK_NUM;
    }
}

void EthernetMaketClient::startConnection(IPAddress ip, uint16_t port, bool nonBlock) {
    dstIP = ip;
    dstPort = port;
    isNonBlocking = nonBlock;
    connectStatus = CONNECT_START;
    startConnectionTime = millis();
}

bool EthernetMaketClient::dataWrite() {
    if (!canWriteIntoConnection(dataSize)) return false;

    if (write(dataPtr, dataSize)) {
        clearData();
        lastActivityTime = millis();
        onSendSuccess();
        return true;
    }

    return false;
}

void EthernetMaketClient::dataWrite(uint8_t* data, uint16_t dataSize, bool override) {
    if (this->dataPtr == nullptr || override) {
        this->dataPtr = data;
        this->dataSize = dataSize;
        activityUpdate();
    } else {
        Serial.println("Data buffer is not empty. Write operation is ignored.");
    }
}

void EthernetMaketClient::clearData() {
    this->dataPtr = nullptr;
    this->dataSize = 0;
}

bool EthernetMaketClient::isDataEmpty() const {
    return this->dataPtr == nullptr;
}

bool EthernetMaketClient::canWriteIntoConnection(size_t size) const {
    return (connectStatus == CONNECT_CONNECTED) && (w5500.getTXFreeSize(_sock) >= size);
}

void EthernetMaketClient::tick() {
    forEach([](EthernetMaketClient &client) {
        client.tick_();
    });
}

void EthernetMaketClient::setOnSendSuccess(const std::function<void()>& callback) {
    onSendSuccess = callback;
}

void EthernetMaketClient::resetOnSendSuccess() {
    onSendSuccess = []() {};
}

void EthernetMaketClient::stop() {
    connectStatus = CONNECT_STOP_START;
}

ConnectionStatusSimple EthernetMaketClient::getConnectionStatus() const {
    return connectStatus;
}

void EthernetMaketClient::activityUpdate() {
    lastActivityTime = millis();
}
